"""
This module shows an example of encrypt and decrypt using RSA algorithm.
"""
import os
from pathlib import Path

from Crypto.PublicKey import RSA
from Crypto.Random import get_random_bytes
from Crypto.Cipher import AES, PKCS1_OAEP
import pycryptex


def encrypt_data(clear_data: bytes, public_key: str) -> list:
    """
    This example do the following:
    - read the public key
    - create a random AES key of 128 bits
    - encrypt the AES key with the public key
    - encrypt data with the AES key (as the traditional symmetric algorithm does)
    - then write in a file the following information:
      - encrypted AES key (first 16 bits)
      - nonce utilized from the AES cypher
      - tag of the AES cypher
      - AES cypher bytes

    :param clear_data: list of bytes to encrypt
    :param public_key: RSA key used for encryption
    :return: encrypted bytes
    """
    # data = "I met aliens in UFO. Here is the map.".encode("utf-8")
    # file_out = open("encrypted_data.bin", "wb")
    encbytes_out = list()
    recipient_key = RSA.import_key(open(public_key).read())
    session_key = get_random_bytes(32)

    # Encrypt the session key with the public RSA key
    cipher_rsa = PKCS1_OAEP.new(recipient_key)
    enc_session_key = cipher_rsa.encrypt(session_key)

    # Encrypt the data with the AES session key
    cipher_aes = AES.new(session_key, AES.MODE_EAX)
    ciphertext, tag = cipher_aes.encrypt_and_digest(clear_data)
    # for x in (enc_session_key, cipher_aes.nonce, tag, ciphertext):
    #     print(f"x: {x}\n")
    for x in (enc_session_key, cipher_aes.nonce, tag, ciphertext):
        encbytes_out.append(x)
    return encbytes_out


def encrypt_file(file: str, public_key: str, remove=False) -> str:
    """
    Encrypt the file and create a new file appending .enc.

    :param file: file to encrypt
    :param public_key: RSA key used for encryption
    :param remove: bool to specify if remove original file
    :return: None
    """
    with open(file, 'rb') as byte_reader:
        # Read all bytes
        clear_bytes = byte_reader.read(-1)
    enc_bytes_list = encrypt_data(clear_bytes, public_key)
    enc_filename = "".join((file, ".enc"))
    with open(enc_filename, "wb") as f:
        for b in enc_bytes_list:
            f.write(b)
    if remove:
        os.remove(file)
    return enc_filename

def decrypt_file(file: str, private_key: str, remove=False, passprhase=None):
    """
    Decrypt the file passed as argument and create a new file removing the .enc extension.

    :param file: encrypted file to decrypt
    :param private_key: RSA private key used for decryption
    :param remove: bool to specify if remove the encrypted file
    :return: the name of the file that has been decrypted
    """
    with open(file, 'rb') as byte_reader:
        # Read all bytes
        enc_bytes = byte_reader.read(-1)
    clear_bytes_list = decrypt_data(enc_bytes, private_key, passprhase)
    with open(file[:-4], "wb") as f:
        f.write(clear_bytes_list)
    if remove:
        os.remove(file)
    return file[:-4]


def decrypt_data(enc_data: bytes, private_key: str, passprhase=None) -> bytes:
    """
    Decrypt data doing the following:
    - get encrypted key and nonce and tag + decrypted bytes
    - decrypt the AES encrypted key using the private key
    - use the AES keys and other information to decrypt data

    :param enc_data: encrypted bytes
    :param private_key: RSA private key used for decryption
    :return: list of decrypted bytes
    """

    # load the RSA private key
    private_key = RSA.import_key(open(private_key).read(), passphrase=passprhase)
    # get the single elements from bytes list
    enc_session_key = enc_data[:private_key.size_in_bytes()]
    nonce = enc_data[private_key.size_in_bytes():private_key.size_in_bytes() + 16]
    tag = enc_data[private_key.size_in_bytes() + 16:private_key.size_in_bytes() + 32]
    ciphertext = enc_data[private_key.size_in_bytes() + 32:]

    # Decrypt the session key with the private RSA key
    cipher_rsa = PKCS1_OAEP.new(private_key)
    session_key = cipher_rsa.decrypt(enc_session_key)

    # Decrypt the data with the AES session key
    cipher_aes = AES.new(session_key, AES.MODE_EAX, nonce)
    data = cipher_aes.decrypt_and_verify(ciphertext, tag)
    return data


def create_keys(folder: str, passprhase=None):
    """
    Create a public key and private key pair
    :param folder: directory where to create the my_key and my_key.pub files
    :return: None
    """
    key = RSA.generate(2048)
    private_key = None
    if passprhase:
        private_key = key.export_key(passphrase=passprhase, pkcs=8,
                                     protection="scryptAndAES128-CBC")
    else:
        private_key = key.export_key()

    with open(os.path.join(folder, "my_key"), "wb") as file_out:
        file_out.write(private_key)

    public_key = key.publickey().export_key()
    with open(os.path.join(folder, "my_key.pub"), "wb") as file_out:
        file_out.write(public_key)



import os
import subprocess
import sys
from getpass import getpass
from pycryptex.crypto import rsa
import pycryptex
from os import path
import time
import click
from pycryptex import utils


class Config():
    """
    This class is to pass some configuration trough the commands
    """

    def __init__(self):
        pass


# Decorator that will create a new instance of Config class. The instance is config and the object can be passed
# through the commands
pass_config = click.make_pass_decorator(Config, ensure=True)


@click.group()
@click.version_option(version=None, message="pycryptex CLI application (version: %(version)s)")
@click.option('--verbose', "-v", is_flag=True, help='bool, to specify if needed a verbose mode')
@pass_config
def cli(config, verbose):
    """
    This command is executed before other commands in the same group. The group commands
    are between the main command (pycrypto) and the single commands.

    For instance:
    pycryptex --verbose encrypt test/appway.png

    Using config.verbose is possible to pass verbose from a command to another
    """
    config.verbose = verbose


@cli.command()
@click.argument('file', required=True)
@click.option('--pubkey', default="", help='(optional) specify the RSA public key')
@click.option('--remove', '-r', is_flag=True, help="(optional, bool=False) to indicate if remove or not the "
                                                   "original clear file")
@pass_config
def encrypt(config, file, pubkey, remove):
    """Encrypt a file"""
    try:
        # in case of pubkey is not passed, pycryptex calculates the default path
        if len(pubkey) == 0:
            pubkey = os.path.join(utils.get_home(), '.pycryptex', "my_key.pub")
        # check if the key exists
        if not path.exists(pubkey):
            echo_invalid_key_msg(pubkey, "pubkey")
            return
        # encryption of the file
        f = rsa.encrypt_file(file=file, public_key=pubkey, remove=remove)
        if config.verbose:
            click.echo(click.style(f"pubkey used is: {pubkey}", fg="magenta", bold=False))
            click.echo(click.style(f"config_file loaded: {pycryptex.config_file}", fg="magenta", bold=True))
        click.echo(click.style(f"👍 File encrypted successfully in {f}", fg="green", bold=True))
    except Exception as e:
        click.echo(click.style(f"Houston, help: {e}, {type(e)}", fg="red", bold=True))
        sys.exit(2)


@cli.command()
@click.argument('file', required=True)
@click.option('--privkey', default="", help='(optional) specify the RSA private key')
@click.option('--remove', '-r', is_flag=True, help="(optional, bool=False) passing this option the encrypted file will"
                                                   "be removed")
@click.option('-s', is_flag=True, help="(optional, bool=False) passing this option the decrypted file will"
                                       "be removed (valid only with --pager option)")
@click.option('--pager', '-p', is_flag=True,
              help="(optional, bool=False) to open or not the pager to read decrypted file")
@pass_config
def decrypt(config, file, privkey, remove, s, pager):
    """Decrypt a file"""
    try:
        f = ""
        # in case of pubkey is not passed, pycryptex calculates the default path
        if len(privkey) == 0:
            privkey = os.path.join(utils.get_home(), '.pycryptex', 'my_key')
        if config.verbose:
            click.echo(click.style(f"priv_key used is: {privkey}", fg="magenta", bold=False))
            # check if the key exists
        if not path.exists(privkey):
            echo_invalid_key_msg(privkey, "privkey")
            return
        try:
            f = rsa.decrypt_file(file=file, private_key=privkey, remove=remove)
        except ValueError as e:
            # try again to decrypt passing the passprhase
            passprhase = getpass("Please insert your passprhase: ")
            f = rsa.decrypt_file(file=file, private_key=privkey, remove=remove, passprhase=passprhase)

        # open file in a pager
        if pager:
            # load config file first
            utils.read_config()
            if config.verbose:
                click.echo(click.style(f"config_file loaded: {pycryptex.config_file}", fg="magenta", bold=True))
            exit_code = subprocess.call([pycryptex.config_file['config']['pager'], f])
            if exit_code == 0:
                # if True delete the decrypted file
                time.sleep(pycryptex.config_file['config']['wait_delete_time'])
                if s:
                    os.remove(f)
                    click.echo(click.style("👍 Decrypted file has been removed successfully!", fg="green", bold=True))
                    return
            else:
                click.echo(click.style(f"Houston, we have a problem: the opened subprocess has a return value equal to"
                                       f" {exit_code}", fg="red", bold=True))

        click.echo(click.style(f"👍 File decrypted successfully in {f}!", fg="green", bold=True))
    except ValueError as e:
        click.echo(click.style(f"Houston, we have a problem: it is possible that you use the wrong key file to decrypt "
                               f"the document or that the passprhase is incorrect. \nTry with the private key "
                               f"corresponding to the public key used to encrypt the file.", fg="red", bold=True))
        sys.exit(2)
    except Exception as e:
        click.echo(click.style(f"Houston, we have a problem: {e}, {type(e)}", fg="red", bold=True))
        sys.exit(2)


@cli.command()
@pass_config
def create_keys(config):
    """
    Create a public and private key pair into the
    '$HOME/.pycryptex' folder.
    """
    try:
        # does keys exist in the target folder?
        is_created, pycryptex_folder = utils.create_home_folder()
        if is_created:
            click.echo(click.style(f"👍 .pycryptex folder created in: {pycryptex_folder}", fg="green", bold=False))
        if os.path.exists(os.path.join(pycryptex_folder, 'my_key')) or \
                os.path.exists(os.path.join(pycryptex_folder, 'my_key.pub')):
            click.echo(click.style(
                "[PAY ATTENTION]\n"
                "The standard keys are present into the default .pycryptex folder. If you confirm to proceed and\n"
                "you already have some document encrypted, you will not be able to open them (if you haven't also copied\n"
                "keys in another location!)", fg="red", bold=True))

        answer = input(f"Do you confirm keys creation into {pycryptex_folder}? (y/n)")
        if answer in ('y', 'yes'):
            answer = input(f"To make your password more secure, do you like to add a passprhase? (y/n)")
            passprhase = None
            if answer in ('y', 'yes'):
                passprhase = getpass("Please insert your passprhase: ")
                # passprhase = input("Please insert your passprhase: ")
            # creation of the keys
            rsa.create_keys(pycryptex_folder, passprhase)
            click.echo(
                click.style("New keys created successfully! Now you can use the other commands, happy encryption!",
                            fg="green", bold=True))
        else:
            click.echo("Keys creation aborted by the user")
    except Exception as e:
        click.echo(click.style(f"Houston, we have a problem during the creation of the keys: {e}", fg="red", bold=True))
        sys.exit(2)


@cli.command()
@pass_config
def create_config(config):
    """
    Create the config file in the $HOME/.pycryptex folder if the file doesn't exist
    """
    try:
        if utils.create_config():
            click.echo(click.style(f"👍 pycryptex.toml file created in: "
                                   f"{os.path.join(utils.get_home(), '.pycryptex', 'pycryptex.toml')}", fg="green",
                                   bold=False))
        else:
            click.echo(click.style(f"👍 nothing to do, file "
                                   f"{os.path.join(utils.get_home(), '.pycryptex', 'pycryptex.toml')} already exists!",
                                   fg="magenta", bold=False))
    except Exception as e:
        click.echo(click.style(f"Houston, help: {e}", fg="red", bold=True))
        sys.exit(2)


def echo_invalid_key_msg(missing_path: str, key_name: str):
    click.echo(
        click.style(f"Houston, help: the key is missing in '{missing_path}'", fg="red", bold=False))
    click.echo(f"If you have your own key, pass the --{key_name} argument or, if you need pycryptex create "
               "the keys for you, type:\n"
               "pycryptex create-keys")


if __name__ == '__main__':
    print("main invoked!")
    cli(sys.argv[1:])



"""
Utils module for repetitive jobs.
"""
import os
import sys
from pathlib import Path
import pycryptex
from os import path
import toml
import click


def get_home() -> str:
    return str(Path.home())


def create_home_folder():
    """
    If it does not exist $HOME/.pycryptex folder will be created
    :return:
    """
    pycryptex_folder = os.path.join(get_home(), '.pycryptex')
    if not os.path.exists(pycryptex_folder):
        os.mkdir(pycryptex_folder)
        return True, pycryptex_folder
    return False, pycryptex_folder


def create_config() -> bool:
    """
    If it does not exist $HOME/.pycryptex/pycryptex.toml file will be created
    :return:
    """
    # first check to create $HOME/.pycryptex folder
    create_home_folder()
    pycryptex_config_file = os.path.join(get_home(), '.pycryptex', 'pycryptex.toml')

    if not os.path.exists(pycryptex_config_file):
        with open(pycryptex_config_file, "w") as f:
            f.write("""# Configuration file for pycryptex
[config]
# path to the pager application where to see decrypted file
pager = "vim"
# number of seconds the application will delete a file decrypted passing the -s option flag
wait_delete_time = 0""")
            return True
    return False


def read_config():
    try:
        config_path = os.path.join(get_home(), '.pycryptex', 'pycryptex.toml')
        if path.exists(config_path):
            pycryptex.config_file = toml.load(config_path)
        else:
            pycryptex.config_file = {
                "config": {
                    'pager': 'vim',
                    'wait_delete_time': 0
                }
            }
    except Exception as e:
        click.echo(click.style(f"Houston, we have a problem in read_config: {e}", fg="red", bold=True))
        sys.exit(1)
